//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
//
// Catan-specific stuff
//
// by Christopher Rasmussen, cer@cis.udel.edu
//
// copyright April, 2010
//
//----------------------------------------------------------------------------
//----------------------------------------------------------------------------

#include "Catan.hh"
#include "Catan_Draw.hh"

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------

// lay out initial board

Board::Board()
{
  bool result;

  game = NULL;

  // hexes

  initialize_hexes();
  initialize_terrain_types();

  // keep generating token values until no 6's/8's are adjacent

  do {
    result = initialize_token_values();
  } while (!result);

  // intersections

  initialize_intersections();
  //  print_intersections();

  // window coordinates

  compute_window_vertices(HEX_GRID_X0, HEX_GRID_Y0, HEX_GRID_SIDE);

  initialize_edges();

  //  exit(1);

  // check for sanity

  //  print_terrain_types();
  //  print_token_values();

}

//----------------------------------------------------------------------------

// in window coords.

Intersection *Board::get_closest_intersection(int x, int y)
{
  int inter_row, inter_col;
  int smallest_inter_row, smallest_inter_col;
  Intersection *smallest_I = NULL;
  float dist;
  float smallest_dist = 1000000.0;

  for (inter_row = 0; inter_row < MAX_VERTICAL_INTERSECTIONS; inter_row++) 
    for (inter_col = 0; inter_col < MAX_HORIZONTAL_INTERSECTIONS; inter_col++) 
      if (intersection[inter_row][inter_col]) {
	dist = hypot(x - intersection[inter_row][inter_col]->x_window,
		     y - intersection[inter_row][inter_col]->y_window);
	if (dist < smallest_dist) {
	  smallest_dist = dist;
	  smallest_I = intersection[inter_row][inter_col];
	  smallest_inter_row = inter_row;
	  smallest_inter_col = inter_col;
	}
      }

  //  printf("row %i, col %i\n", smallest_inter_row, smallest_inter_col);

  return smallest_I;
}

//----------------------------------------------------------------------------

// edge from (x1, y1) to (x2, y2); point is (x3, y3)
// following Paul Bourke's approach at http://local.wasp.uwa.edu.au/~pbourke/geometry/pointline/

double Board::point_edge_distance(double x1, double y1, double x2, double y2, double x3, double y3)
{
  double x_diff = x2 - x1;
  double y_diff = y2 - y1;
  double u_num = (x3 - x1) * x_diff + (y3 - y1) * y_diff;
  double u_denom = x_diff*x_diff + y_diff*y_diff;
  double u, x, y, p1_dist, p2_dist, dist;

  if (u_denom)
    u = u_num / u_denom;
  else {
    cout << "0-length line segment!!!\n";
    exit(1);
  }

  if (u < 0 || u > 1) {
    p1_dist = sqrt((x3-x1)*(x3-x1) + (y3-y1)*(y3-y1));
    p2_dist = sqrt((x3-x2)*(x3-x2) + (y3-y2)*(y3-y2));
    if (p1_dist < p2_dist)
      return p1_dist;
    else
      return p2_dist;
  }
  else {
    x = x1 + u*x_diff;
    y = y1 + u*y_diff;
    dist = sqrt((x3-x)*(x3-x) + (y3-y)*(y3-y));
    return dist;
  }
}

//----------------------------------------------------------------------------

// in window coords.

Edge *Board::get_closest_edge(int x, int y)
{
  int i, min_index = -1;
  double dist;
  double min_dist = 10000000000.0;

  for (i = 0; i < edge_list.size(); i++) {
    dist = point_edge_distance(edge_list[i]->I1->x_window,
			       edge_list[i]->I1->y_window,
			       edge_list[i]->I2->x_window,
			       edge_list[i]->I2->y_window, 
			       x, y);
    if (dist < min_dist) {
      min_dist = dist;
      min_index = i;
    }
  }

//   printf("%.0f %.0f -> %.0f %.0f\n", 
// 	 edge_list[min_index]->I1->x_window,
// 	 edge_list[min_index]->I1->y_window,
// 	 edge_list[min_index]->I2->x_window,
// 	 edge_list[min_index]->I2->y_window);

  if (min_index >= 0)
    return edge_list[min_index];
  else
    return NULL;

  /*
  int row, col;
  Intersection *smallest_I = NULL;
  float dist;
  float smallest_dist = 1000000.0;

  for (row = 0; row < MAX_VERTICAL_INTERSECTIONS; row++) 
    for (col = 0; col < MAX_HORIZONTAL_INTERSECTIONS; col++) 
      if (intersection[row][col]) {
	dist = hypot(x - intersection[row][col]->x_window,
		     y - intersection[row][col]->y_window);
	if (dist < smallest_dist) {
	  smallest_dist = dist;
	  smallest_I = intersection[row][col];
	}
      }

  return smallest_I;
  */
}

//----------------------------------------------------------------------------

// compute window vertices for each hex in grid
// (0, 0) is upper-left corner of grid

void Board::compute_window_vertices(float x0, float y0, float side_length)
{
  int row, col;
  float x, y, x_off;

  for (row = 0, x_off = 0.0; row < MAX_HEXES_PER_SIDE; row++, x_off -= 0.5) 
    for (col = 0; col < MAX_HEXES_PER_SIDE; col++) 
      if (hex[row][col]) {

	// center of hex in window coordinates

	x = x0 + (x_off + (float) col) * HEX_WIDTH_FACTOR * side_length;
	y = y0 - (float) row * side_length * (HEX_HEIGHT_FACTOR - HEX_NOTCH_FACTOR);

	// get vertex coords from center

	hex[row][col]->compute_window_vertices(x, y, side_length);
      }
}

//----------------------------------------------------------------------------

// convert mouse coordinates to hex (row, col)
// returns true if mouse over board hex, false otherwise

bool Board::window_point_inside_which_hex(float x, float y, int & which_row, int & which_col)
{
  int row, col;

  for (row = 0; row < MAX_HEXES_PER_SIDE; row++) 
    for (col = 0; col < MAX_HEXES_PER_SIDE; col++)
      if (hex[row][col] && hex[row][col]->window_point_inside(x, y)) {
	which_row = row;
	which_col = col;
	return true;
      }

  return false;
}

//----------------------------------------------------------------------------

void Board::clear_hex_highlights()
{
  int row, col;

  for (row = 0; row < MAX_HEXES_PER_SIDE; row++) 
    for (col = 0; col < MAX_HEXES_PER_SIDE; col++)
      if (hex[row][col])
	hex[row][col]->draw_highlighted = false;
}

//----------------------------------------------------------------------------

void Board::draw()
{
  int row, col, inter_row, inter_col, i;

  // ocean background done in display() proper

  // background colors

  for (row = 0; row < MAX_HEXES_PER_SIDE; row++) 
    for (col = 0; col < MAX_HEXES_PER_SIDE; col++)
      if (hex[row][col]) 
	hex[row][col]->draw_terrain();

  // token values & outline

  for (row = 0; row < MAX_HEXES_PER_SIDE; row++) 
    for (col = 0; col < MAX_HEXES_PER_SIDE; col++) 
      if (hex[row][col]) {
	hex[row][col]->draw_token_value();
	//	hex[row][col]->draw_outline(0, 0, 0, 1);
	hex[row][col]->draw_outline(153, 153, 179, 7);
	if (hex[row][col]->robber)
	  draw_robber(hex[row][col]->xc, hex[row][col]->yc, hex[row][col]->game->robber_sprite);
      }


  // highlights

  for (row = 0; row < MAX_HEXES_PER_SIDE; row++) 
    for (col = 0; col < MAX_HEXES_PER_SIDE; col++) 
      if (hex[row][col] && hex[row][col]->draw_highlighted)
	//	hex[row][col]->draw_outline(255, 0, 255, 5, 0.9);
	hex[row][col]->draw_outline(255, 0, 255, 5, 0.8);
  
  // roads

  float x, y;

  for (i = 0; i < edge_list.size(); i++)
    if (edge_list[i]->road) {

      draw_line(edge_list[i]->I1->x_window, edge_list[i]->I1->y_window,
		edge_list[i]->I2->x_window, edge_list[i]->I2->y_window,
		PLAYER_R[edge_list[i]->owned_by], PLAYER_G[edge_list[i]->owned_by], PLAYER_B[edge_list[i]->owned_by],
		ROAD_THICKNESS);

      //      draw_road();
    }

  // intersections

  for (inter_row = 0; inter_row < MAX_VERTICAL_INTERSECTIONS; inter_row++) 
    for (inter_col = 0; inter_col < MAX_HORIZONTAL_INTERSECTIONS; inter_col++) 
      if (intersection[inter_row][inter_col])
	intersection[inter_row][inter_col]->draw();



  // ports

  if (!game)
    return;

  int w = game->trade_wildcard_sprite->imwidth;
  int h = game->trade_wildcard_sprite->imheight;

  glEnable(GL_TEXTURE_2D);
  glEnable(GL_BLEND);

  for (i = 0; i < NUM_WILDCARD_PORT_ICONS; i++)
    game->trade_wildcard_sprite->draw(WILDCARD_PORT_X[i], WILDCARD_PORT_Y[i], w, h, 0, 1, 1, 1);

  for (i = 0; i < NUM_RESOURCES; i++)
    game->resource_sprite[i]->draw_sub_sprite(RESOURCE_PORT_X[i], RESOURCE_PORT_Y[i],
					      RESOURCE_SUB_SDELTA*(double)game->resource_sprite[i]->imwidth, 
					      RESOURCE_SUB_TDELTA*(double)game->resource_sprite[i]->imheight, 
					      0.0,
					      RESOURCE_SUB_S, RESOURCE_SUB_T, RESOURCE_SUB_SDELTA, RESOURCE_SUB_TDELTA, 
					      1, 1, 1);

  glDisable(GL_TEXTURE_2D);
  glDisable(GL_BLEND);

  //  if (hex_highlighted) {

    // highlight adjacent hexes


    /*
    // highlight this hex

    hex[highlight_row][highlight_col]->draw_outline(SELECTED_HEX_R, SELECTED_HEX_G, SELECTED_HEX_B, 5);

    // highlight adjacent intersections

    for (int i = 0; i < NUM_INTERSECTIONS_PER_HEX; i++) 
      hex[highlight_row][highlight_col]->inter[i]->draw_highlighted(0, 255, 0, 8);

  }

  if (inter_highlighted) {

    intersection[highlight_inter_row][highlight_inter_col]->draw_highlighted(0, 0, 255, 16);

  }
  */

}

//----------------------------------------------------------------------------

void Board::initialize_hexes()
{
  int row, col;

  for (row = 0; row < MAX_HEXES_PER_SIDE; row++) 
    for (col = 0; col < MAX_HEXES_PER_SIDE; col++) {
      if (is_board_hex(row, col)) 
	hex[row][col] = new Hex(row, col, this);
      else
	hex[row][col] = NULL;
    }
}

//----------------------------------------------------------------------------

void Board::initialize_intersections()
{
  int inter_row, inter_col;

  for (inter_row = 0; inter_row < MAX_VERTICAL_INTERSECTIONS; inter_row++) 
    for (inter_col = 0; inter_col < MAX_HORIZONTAL_INTERSECTIONS; inter_col++) {
      if (is_board_intersection(inter_row, inter_col)) 
	intersection[inter_row][inter_col] = new Intersection(inter_row, inter_col, this);
      else
	intersection[inter_row][inter_col] = NULL;
    }
}

//----------------------------------------------------------------------------

bool Board::edge_adjacent_to_intersection(Edge *E, Intersection *I)
{
  return (E && I && (E->I1 == I || E->I2 == I));
}

//----------------------------------------------------------------------------

bool Board::edge_adjacent_to_edge(Edge *E1, Edge *E2)
{
  return (E1 && E2 && E1 != E2 && 
	  (E1->I1 == E2->I1 || E1->I2 == E2->I1) ||
	  (E1->I1 == E2->I2 || E1->I2 == E2->I2));
}

//----------------------------------------------------------------------------

// have we put this edge in the list already?

bool Board::is_new_edge(Intersection *I1, Intersection *I2)
{
  int i;

  for (i = 0; i < edge_list.size(); i++) 
    if (edge_list[i]->I1 == I1 && edge_list[i]->I2 == I2 ||
	edge_list[i]->I1 == I2 && edge_list[i]->I2 == I1)
      return false;

  return true;
}

//----------------------------------------------------------------------------

void Board::initialize_edges()
{
  int inter_row, inter_col;
  int inter_row2, inter_col2;
  Intersection *I1, *I2;

  edge_list.clear();

  // iterate over intersections

  for (inter_row = 0; inter_row < MAX_VERTICAL_INTERSECTIONS; inter_row++) 
    for (inter_col = 0; inter_col < MAX_HORIZONTAL_INTERSECTIONS; inter_col++) {
      I1 = intersection[inter_row][inter_col];
      if (I1) {

	// iterate over intersections adjacent to this one

	for (inter_row2 = 0; inter_row2 < MAX_VERTICAL_INTERSECTIONS; inter_row2++) 
	  for (inter_col2 = 0; inter_col2 < MAX_HORIZONTAL_INTERSECTIONS; inter_col2++) {
	    I2 = intersection[inter_row2][inter_col2];
	    if (I1->is_adjacent(I2)) {
	      //	      printf("Intersections at %i, %i and %i, %i are adjacent\n", I1->row, I1->col, I2->row, I2->col);
	      if (is_new_edge(I1, I2)) {
		//		printf("%i: New edge--adding!\n", edge_list.size());
		edge_list.push_back(new Edge(I1, I2));
	      }
	      //	      else
	      //		printf("already added\n");
	    }
	  }
      }
    }
}

//----------------------------------------------------------------------------

void Board::print_intersections()
{
  int inter_row, inter_col;

  cout << "  ";
  for (inter_col = 0; inter_col < MAX_HORIZONTAL_INTERSECTIONS; inter_col++) 
    printf("%X", inter_col);
  cout << endl;

  for (inter_row = 0; inter_row < MAX_VERTICAL_INTERSECTIONS; inter_row++) {
    cout << inter_row << " ";
    for (inter_col = 0; inter_col < MAX_HORIZONTAL_INTERSECTIONS; inter_col++) {
      if (is_board_intersection(inter_row, inter_col)) {
	if (intersection[inter_row][inter_col]->building == PIECE_SETTLEMENT)
	  cout << "s";
	else if (intersection[inter_row][inter_col]->building == PIECE_CITY)
	  cout << "c";
	else
	  cout << "+";
      }
      else
	cout << ".";
    }
    cout << "\n";
  }
}

//----------------------------------------------------------------------------

// make each board hex one of { hills, pasture, mountains, fields, forest, desert }

// this actually calls hex constructor, so it should be called first

void Board::initialize_terrain_types()
{
  int row, col;

  // how many hexes of each terrain type we start with

  int remaining_terrains[NUM_TERRAINS] = { NUM_HILLS_HEXES, 
					   NUM_PASTURE_HEXES, 
					   NUM_MOUNTAINS_HEXES, 
					   NUM_FIELDS_HEXES, 
					   NUM_FOREST_HEXES, 
					   NUM_DESERT_HEXES };

  // visit each hex, pick a random terrain, and decrement the number available

  for (row = 0; row < MAX_HEXES_PER_SIDE; row++) 
    for (col = 0; col < MAX_HEXES_PER_SIDE; col++) 
      if (hex[row][col])
	hex[row][col]->terrain = pick_random_available(remaining_terrains, NUM_TERRAINS);
}

//----------------------------------------------------------------------------

// put a token with a value between 2 and 12, inclusive on each non-desert board hex

// afterwards, checks if any 6's or 8's are adjacent to one another.  if so, return false--board is
// not good.  else return true--board is ready for play

bool Board::initialize_token_values()
{
  int row, col;

  // how many tokens of each value we start with

  int remaining_tokens[NUM_TOKEN_VALUES] = { NUM_0s, NUM_1s, NUM_2s, NUM_3s, NUM_4s, NUM_5s, NUM_6s, 
					     NUM_7s, NUM_8s, NUM_9s, NUM_10s, NUM_11s, NUM_12s };

  // visit each hex, pick a random token value, and decrement the number available

  for (row = 0; row < MAX_HEXES_PER_SIDE; row++) 
    for (col = 0; col < MAX_HEXES_PER_SIDE; col++) 
      if (hex[row][col]) {
	if (hex[row][col]->terrain == TERRAIN_DESERT) {
	  hex[row][col]->token_value = ROBBER_TOKEN;
	  hex[row][col]->robber = true;
	}
	else {
	  hex[row][col]->token_value = pick_random_available(remaining_tokens, NUM_TOKEN_VALUES); 
	  if (hex[row][col]->is_special() && hex[row][col]->is_adjacent_to_special())
	    return false;
	}
      }

  return true;
}

//----------------------------------------------------------------------------

// terrain/token arrays store how many of each type still remain to be chosen.
// this function should NOT be called if pile[] is all 0's, because it will not detect this case

int Board::pick_random_available(int pile[], int num_piles)
{
  int which_type;
  int i;

  while (1) {
    which_type = uniform_random_int(0, num_piles - 1);
    if (pile[which_type] > 0) {
      pile[which_type]--;
      return which_type;
    }
  }
}

//----------------------------------------------------------------------------

// iterate through all hexes and print intersection adjacencies for each one

void Board::print_all_intersection_adjacencies()
{
  int row, col;

  cout << "-------------\nAll intersection adjacencies\n-------------\n";

  for (row = 0; row < MAX_HEXES_PER_SIDE; row++) 
    for (col = 0; col < MAX_HEXES_PER_SIDE; col++) {
      if (hex[row][col]) 
	hex[row][col]->print_intersection_adjacencies();
      cout << "\n\n";
    }
}

//----------------------------------------------------------------------------

// iterate through all hexes and print adjacencies for each one

void Board::print_all_hex_adjacencies()
{
  int row, col;

  cout << "-------------\nAll adjacencies\n-------------\n";

  for (row = 0; row < MAX_HEXES_PER_SIDE; row++) 
    for (col = 0; col < MAX_HEXES_PER_SIDE; col++) {
      if (hex[row][col]) 
	hex[row][col]->print_hex_adjacencies();
      cout << "\n\n";
  }
}

//----------------------------------------------------------------------------

// what terrain is at each hex?

void Board::print_terrain_types(bool debug)
{
  int row, col;

  cout << "-------------\nTerrain types\n-------------\n";

  for (row = 0; row < MAX_HEXES_PER_SIDE; row++) {
    for (col = 0; col < MAX_HEXES_PER_SIDE; col++) {
      if (is_board_hex(row, col)) {
	if (debug) 
	  cout << hex[row][col]->terrain;
	else
	  cout << TERRAIN_CODE[hex[row][col]->terrain];
      }
      else
	cout << ".";
    }
    cout << "\n";
  }
}

//----------------------------------------------------------------------------

// what token value is at each hex?

void Board::print_token_values()
{
  int row, col;

  cout << "-------------\nToken values\n-------------\n";

  for (row = 0; row < MAX_HEXES_PER_SIDE; row++) {
    for (col = 0; col < MAX_HEXES_PER_SIDE; col++) {
      if (is_board_hex(row, col)) {
	printf("%X", hex[row][col]->token_value);
      }
      else
	cout << ".";
    }
    cout << "\n";
  }
}

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------


