//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
//
// OpenGL Sprite class v 1.1
//
// by Christopher Rasmussen, cer@cis.udel.edu
//
// CISC 440/660, September, 2008
// modified April, 2010
//
// copyright 2010
//
//----------------------------------------------------------------------------
//----------------------------------------------------------------------------

#include "Sprite.hh"

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------

// globals :( ... hey, opengl does it

vector <string> texfilename;
vector <int> texname;
map <int, int> texname_index;  // mapping from texname int to load order index
vector <RGB> trans_rgb;
int *texim_w, *texim_h;        // holds "natural" dimensions of texture images in load order

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------

Sprite::Sprite(int itn)
{ 
  texture_number = itn; 

  // behold the mighty power of STL!

  imwidth = texim_w[texname_index[texture_number]];
  imheight = texim_h[texname_index[texture_number]];
  
  s = t = 0.0; 
  sdelta = tdelta = 1.0; 
}

//----------------------------------------------------------------------------

// display the sprite centered at (x, y) and rotated theta degrees counterclockwise
// r, g, b "tints" texture with specified color

void Sprite::draw(float x, float y, float w, float h, float theta, float r, float g, float b)
{
  glColor3f(r, g, b);

  glBindTexture(GL_TEXTURE_2D, texture_number);  

  glPushMatrix();

  glTranslatef(x, y, 0);
  glRotatef(theta, 0, 0, 1);

  glBegin(GL_QUADS);
    glTexCoord2f(s, t);                   glVertex3f(-0.5*w,  0.5*h, 0.0);   // bottom left
    glTexCoord2f(s, t + tdelta);          glVertex3f(-0.5*w, -0.5*h, 0.0);   // top left
    glTexCoord2f(s + sdelta, t + tdelta); glVertex3f( 0.5*w, -0.5*h, 0.0);   // top right
    glTexCoord2f(s + sdelta, t);          glVertex3f( 0.5*w,  0.5*h, 0.0);   // bottom right
  glEnd();

  glPopMatrix();

  glColor3f(1,1,1);
}

//----------------------------------------------------------------------------

// draw "subwindow" of full sprite texture

void Sprite::draw_sub_sprite(float x, float y, float w, float h, float theta, 
			     float stemp, float ttemp, float sdeltatemp, float tdeltatemp,
			     float r, float g, float b)
{
  if (stemp < 0.0 || stemp > 1.0 || ttemp < 0.0 || ttemp > 1.0) {
    printf("bad subsprite parameters on s or t\n");
    return;
  }
  if ((stemp + sdeltatemp) < 0.0 || (stemp + sdeltatemp) > 1.0 || (ttemp + tdeltatemp) < 0.0 || (ttemp + tdeltatemp) > 1.0) {
    printf("bad subsprite parameters on sdelta (%f) or tdelta (%f)\n", sdeltatemp, tdeltatemp);
    return;
  }

  s = stemp;
  t = ttemp;
  sdelta = sdeltatemp;
  tdelta = tdeltatemp;

  draw(x, y, w, h, theta, r, g, b);

  s = t = 0.0;
  sdelta = tdelta = 1.0;
}

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------

MultiSprite::MultiSprite(int icols, int irows, int itn) : Sprite(itn)
{ 
  num_cols = icols; 
  num_rows = irows; 
  sd = 1.0 / (float) num_cols; 
  td = 1.0 / (float) num_rows; 
}

//----------------------------------------------------------------------------

// col, row sets which frame is getting drawn.
// see parameter explanations above

void MultiSprite::draw(int col, int row, float x, float y, float w, float h, float theta, float r, float g, float b)
{
  if (col < 0 || col > num_cols - 1 || row < 0 || row > num_rows - 1)
    return;

  s = sd * (float) col;
  t = td * (float) row;
  sdelta = sd;
  tdelta = td;
  ((Sprite *) this)->draw(x, y, w, h, theta, r, g, b);
  
  s = t = 0.0; sdelta = tdelta = 1.0;
}

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------

void clear_sprites()
{
  texfilename.clear();
  texname.clear();
  trans_rgb.clear();
  texname_index.clear();
}

//----------------------------------------------------------------------------

void add_sprite_info(string filename, int id, RGB & transparency_rgb)
{
  texfilename.push_back(filename);
  texname.push_back(id);
  trans_rgb.push_back(transparency_rgb);
  texname_index[id] = texname.size() - 1;
}

//----------------------------------------------------------------------------

// read in sprite textures and associate them with texture "names" (positive integers)
// texfilename and texname should be same length, and neither should have any repeats.

void finalize_sprites()
{
  int i;
  unsigned char **texim;

  texim = (unsigned char **) calloc(texfilename.size(), sizeof(unsigned char *));
  texim_w = (int *) calloc(texfilename.size(), sizeof(int));
  texim_h = (int *) calloc(texfilename.size(), sizeof(int));

  for (i = 0; i < texfilename.size(); i++) {

    texim[i] = read_PPM(texfilename[i].c_str(), &texim_w[i], &texim_h[i], trans_rgb[i]);

    printf("loading %s: w = %i, h = %i\n", texfilename[i].c_str(), texim_w[i], texim_h[i]);

    glBindTexture(GL_TEXTURE_2D, texname[i]);

    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);

    // not sure if this is making a copy of texim[i] (probably--in which case it should be de-allocated)
    // or just copying the pointer

    glTexImage2D(GL_TEXTURE_2D, 0, 4, texim_w[i], texim_h[i], 0, GL_RGBA, 
		 GL_UNSIGNED_BYTE, texim[i]);
  }
}

//----------------------------------------------------------------------------

// read in a PPM image file in order to bind it to a texture

// (trans_r, trans_g, trans_b) sets color key for which pixels should be treated as transparent.
// if you want none to be transparent, pass in impossible values like (-1, -1, -1)

unsigned char *read_PPM(const char *filename, int *width, int *height, RGB & transparent)
{
  int i, j, w, h;
  FILE *fp;
  int r, g, b, t;
  bool iscomment;
  unsigned char ur, ug, ub;
  char imtype[3];
  char str[80];
  char c;
  unsigned char *im;
  int bpp;

  // get size info

  fp = fopen(filename, "rb");
  if (!fp) {
    printf("read_PPM(): no such file");
    exit(1);
  }

  fscanf(fp, "%s\n", &imtype);

  // attempt to eat comments

  do {
    iscomment = false;
    c = fgetc(fp);
    ungetc(c, fp);
    if (c == '#') {
      iscomment = true;
      fgets (str, 79, fp);
    }
  } while (iscomment);

  // read image dimensions

  fscanf(fp, "%i %i\n255\n", &w, &h);    
  *width = w;
  *height = h;

  // allocate image

  bpp = 4;

  im = (unsigned char *) calloc(w * h * bpp, sizeof(unsigned char));

  // actually read it in

  // ascii

  if (!strcmp(imtype, "P3")) {

    for (j = 0, t = 0; j < h; j++) {
      for (i = 0; i < w; i++, t += bpp) {
	fscanf(fp, "%i %i %i ", &r, &g, &b);
	im[t] = (unsigned char) r;
	im[t+1] = (unsigned char) g;
	im[t+2] = (unsigned char) b;
	if (r == transparent.r && g == transparent.g && b == transparent.b)   // this is where transparency is set
	  im[t+3] = (unsigned char) 0;
	else
	  im[t+3] = (unsigned char) 255;
      }
    }
    fscanf(fp, "\n");
  }

  // binary

  else if (!strcmp(imtype, "P6")) {
    
    for (j = 0, t = 0; j < h; j++) {
      for (i = 0; i < w; i++, t += bpp) {
	fscanf(fp, "%c%c%c", &ur, &ug, &ub);
	im[t] = (unsigned char) ur;
	im[t+1] = (unsigned char) ug;
	im[t+2] = (unsigned char) ub;
	if (ur == transparent.r && ug == transparent.g && ub == transparent.b)  // this is where transparency is set
	  im[t+3] = (unsigned char) 0;
	else
	  im[t+3] = (unsigned char) 255;
      }
    }
  }

  // unknown

  else {
    printf("unrecognized ppm file type");
    exit(1);
  }

  // finish up

  fclose(fp);

  return im;
}

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
